/* mfaim
 * Copyright (c) 1999,2000 Josh Myer <josh@joshisanerd.com>
 * console-mode AIM client
 *
 * mfaim is based on naim;
 *  naim is Copyright (c) 1998, 1999 Daniel Reed <n@ml.org>
 * 
 * mfaim also uses libfaim,
 *  FAIM is Copyright 1998 Adam Fritzler <afritz@iname.com>
 *  FAIM uses code Copyright 1998 Brock Wilcox <awwaiid@iname.com>
 */

#include <mfaim.h>

extern WINDOW	*win_input;
extern buddywin_t	*curbwin, nullbwin;
extern faimconf_t	faimconf;
extern struct aim_conn_t	*nfaim_conn;
extern struct aim_session_t mfaim_session;
extern char *faimrcfile;

int input_over = 0;

void conio_exit(char *params) {
  if (params != NULL) {
    buddywin_t *bwin = curbwin;
   
    while( (bwin = bwin->next) != curbwin)
      if(bwin != &nullbwin) 
	conio_sendim(bwin->winname,params);      
    
  }
  aim_logoff(&mfaim_session);
  wshutitdown();
  /* So we don't quite conform to the structured programming ideal.
   * Oh well.  JMM */
  exit(0);
}

void conio_info(char *SN) {
  if (SN == NULL) {
    if (curbwin == &nullbwin) 
      SN = faimconf.screenname;
    else
      SN = curbwin->winname; /* SN's scope is shorter then curbwin's. */ 
  }
  aim_getinfo(&mfaim_session, nfaim_conn, SN, AIM_GETINFO_GENERALINFO);
  return;
}

void conio_closewin(char *SN) {
  buddywin_t	*bwin, *bnext, *bprev;

  if (SN == NULL) {
    if ((bwin = curbwin) == &nullbwin) {
      printf("You cannot close the status window.\n");
      return;
    }
  } else {
    if ((bwin = wgetwin(SN)) == &nullbwin) {
      printf("No window is open for %s.\n", SN);
      return;
    }
  }

  bnext = bwin->next;
  bprev = bwin->prev;

  bprev->next = bnext;
  bnext->prev = bprev;

  curbwin = bnext;

  wremove(bwin);
  wbuddy_update();

  touchwin(curbwin->win);
  return;
}

void conio_openwin(char *SN) {
  
  if (strlen(SN) > MAXWLEN) {
    printf("Screen names must be %d characters or shorter.\n", MAXWLEN);
  } else {
    buddywin_t *bwin = NULL;
    
    bwin = wgetwin(SN);
    
    if (bwin == &nullbwin) {
      bnewwin(SN);
      bwin = wgetwin(SN);
      bwin->offline = 0;
      bwin->notbuddy = 1;
      wprintf(bwin->win, IMWIN_STATUS, 1,"Non-buddy window created.\n");
    } else {
      if (bwin->win != nullbwin.win) 
	printf("There is already a window open for %s.\n", SN);
      else {
	bwin->win = bnewwindow();
	printf("Opened window for %s.\n", SN);
      }
    }
  }
  wbuddy_update();
  return;
}


void	conio_sendim(char *SN, char *msg) {
  buddywin_t *bwin;

  if(SN == NULL)
    return;

  if (strlen(SN) > MAXSNLEN) {
    printf("Screen names must be %d characters or shorter.\n", MAXSNLEN);
    return;
  }
  if (msg == NULL) {
    aim_getinfo(&mfaim_session, nfaim_conn, SN, AIM_GETINFO_GENERALINFO);
    return;
  }

  mfaim_send_im(SN, msg); /* send it first. hopefully it has time to
			     get out if we die below */

  bwin = wgetwin(SN);

  if(bwin == &nullbwin)
    mfaim_dprintf(curbwin->win, IMWIN_TEXT_U, "-> *<B>%s</B>* %s<br>",  SN, msg);
  else {
    if (secs_getvar_int("timestamps") & 0x02)
      mfaim_dprintf(curbwin->win, IMWIN_TEXT_U, "<B>%s</B>: %s<br>", mfaim_session.logininfo.screen_name, msg);
    else
      hwprintf(curbwin->win, IMWIN_TEXT_U, "<B>%s</B>: %s<br>", mfaim_session.logininfo.screen_name, msg);
  }

  logim(mfaim_session.logininfo.screen_name, SN, msg);
  return;
}

void	conio_sendawayim(char *SN, char *msg) {
  buddywin_t *bwin;

  if(SN == NULL)
    return;

  if (strlen(SN) > MAXSNLEN) {
    printf("Screen names must be %d characters or shorter.\n", MAXSNLEN);
    return;
  }
  if (msg == NULL) {
    aim_getinfo(&mfaim_session, nfaim_conn, SN, AIM_GETINFO_GENERALINFO);
    return;
  }

  mfaim_send_im_away(SN, msg); /* send it first. hopefully it has time to
			     get out if we die below */

  bwin = wgetwin(SN);

  if(bwin == &nullbwin)
    mfaim_dprintf(curbwin->win, IMWIN_TEXT_U, "(AWAY)-> *<B>%s</B>* %s<br>",  SN, msg);
  else
    hwprintf(curbwin->win, IMWIN_TEXT_U, "(AWAY)<B>%s</B>: %s<br>", mfaim_session.logininfo.screen_name, msg);

  logim(mfaim_session.logininfo.screen_name, SN, msg);
  return;
}


void	conio_sendchatim(char *SN, char *msg) {

  if(SN == NULL)
    return;

  if (strlen(SN) > MAXCRLEN) {
    printf("Room names must be %d characters or shorter.\n", MAXCRLEN);
    return;
  }
  if (msg == NULL)
    return;

  /* we don't have to log or print here -- AIM sends it back to us! */

  if (aim_chat_getconn(&mfaim_session, SN) == NULL)
    printf("Not connected to room %s\n", SN);
  else
    aim_chat_send_im(&mfaim_session, aim_chat_getconn(&mfaim_session, SN), msg);
  return;
}

void	conio_jump(char *SN) {
  buddywin_t	*bwin = NULL;
  char snbuf[MAXWLEN+1];
  if (SN == NULL) {
#if USE_STACK
    sn_queue_deq(snbuf);
    if(snbuf != NULL)
      conio_jump(snbuf);
#endif
    return;
  } else {
    if ((bwin = wgetwin(SN)) == &nullbwin) 
      if(strcasecmp(nullbwin.winname, bwin->winname)!=0){
	printf("No such buddy %s.\n", SN);
	return;
      }
    if (bwin != curbwin) {
      curbwin = bwin;
      wbuddy_update();
      touchwin(curbwin->win);
    }
  }
  return;
}

void conio_jump_j(char *SN) {
  buddywin_t *bwin = NULL;
  char snbuf[MAXWLEN+1];
  if (SN == NULL) {
#if USE_STACK
    sn_queue_deq(snbuf);
    if(snbuf != NULL)
      conio_jump(snbuf);
#endif
    return;
  } else {
    if ((bwin = wgetwin_u(SN)) == NULL) {
      return;
    }
    curbwin = bwin;
    wbuddy_update();
    touchwin(curbwin->win);
  }
  return;
}

void conio_reload(char *file) {
  char buf[256];

  killconf(&faimconf);
  if (file == NULL) {
    if (faimrcfile)
      file = faimrcfile;
    else
      {
	snprintf(buf, sizeof(buf), "%s/.faimrc", getenv("HOME"));
	file = buf;
      }
  }
  printf("Re-reading %s\n", file);
  mfaim_read_config(file);
  wsetp();
  mfaim_budprof(nfaim_conn);
}

void conio_away(char *awaymsg) {
  if(awaymsg) {
    secs_setvar("awaymsg", awaymsg);
    setaway();
  } else {
    if(nullbwin.away)
      unsetaway();
    else
      setaway();
  }
  return;
}

int	conio_handlecmd(char *buf) {
  char	line[MAXCMDLEN+1];
  char	*cmd, *parms;

  assert(buf != NULL);
  strncpy(line, buf, sizeof(line)-1);
  line[sizeof(line)-1] = 0;
  cmd = atom(line);
  parms = firstwhite(line);

  if (script_doalias(cmd, parms) == 1)
    return(1);
  if (cmd){
    if (strcasecmp(cmd, "EVAL") == 0) {
      secs_script_parse(parms);
    } else if ((strcasecmp(cmd, "IM") == 0)
	       || (strcasecmp(cmd, "M") == 0)
	       || (strcasecmp(cmd, "MSG") == 0)) {
      conio_sendim(atom(parms), firstwhite(parms));
    } else if (strcasecmp(cmd, "CLEAR") == 0) {
      werase(curbwin->win);
    } else if (strcasecmp(cmd, "RELOAD") == 0) {
      conio_reload(parms);
    } else if (strcasecmp(cmd, "AWAY") == 0) {
      conio_away(parms);
    } else if (strcasecmp(cmd, "SET") == 0) {
      set_setvar(parms);
    } else if ((strcasecmp(cmd, "OPEN") == 0)
	       || (strcasecmp(cmd, "WINDOW") == 0)) {
      conio_openwin(parms);
    } else if ((strcasecmp(cmd, "CLOSE") == 0)
	       || (strcasecmp(cmd, "ENDWIN") == 0)
	       || (strcasecmp(cmd, "PART") == 0)) {    
      conio_closewin(parms);
    } else if ((strcasecmp(cmd, "HELP") == 0)
	       || (strcasecmp(cmd, "ABOUT") == 0)) {
      help_printhelp(parms);
    } else if (strcasecmp(cmd, "J") == 0) {
      conio_jump_j(parms);
    } else if (strcasecmp(cmd, "JUMP") == 0) {
      conio_jump(parms);
    } else if ((strcasecmp(cmd, "QUIT") == 0)
	       || (strcasecmp(cmd, "EXIT") == 0)) {
      conio_exit(parms);
    } else if ((strcasecmp(cmd, "INFO") == 0)
	       || (strcasecmp(cmd, "WHOIS") == 0)) {
      conio_info(parms);
    } else if (strcasecmp(cmd, "WHO") == 0) {
      conio_who(parms);
    } else if ((strcasecmp(cmd, "JOIN") == 0)
	       || (strcasecmp(cmd, "JOINROOM") == 0)) {
      conio_joinroom(parms);
    } else if (strcasecmp(cmd, "INVITE") == 0) {
      conio_invite(parms);
    } else if (strcasecmp(cmd, "AMSG") == 0) {
      conio_sendawayim(atom(parms), firstwhite(parms));
    } else {
      return(0);
    }
  }
  return(1);
}
int      (*secs_client_cmdhandler)(char *) = conio_handlecmd;

void conio_invite(char *parms) {
  if (parms)
    aim_chat_invite(&mfaim_session, nfaim_conn, parms, "Join me over here yonder...", 0x0004, curbwin->winname, 0x0000);
  else
    printf("invite needs a screen name\n");
  return;
}

void conio_joinroom(char *parms) {
  if (parms)
    aim_chat_join(&mfaim_session, nfaim_conn, 0x0004, parms);
  else
    printf("joinroom needs a room name\n");
  return;
}

void conio_who(char *list) {
  if (list == NULL) {
    buddywin_t *bwin;
    bwin = curbwin;
    wprintf(curbwin->win, STAT_TEXT, 1, "%s", "Buddies Currently Online:");
    do {
      if(bwin != &nullbwin)
	wprintf(curbwin->win, STAT_TEXT, 1, "%s", bwin->winname);
      if(bwin->next != curbwin)
	wprintf(curbwin->win, STAT_TEXT, 1, ", ");
    } while((bwin = bwin->next) != curbwin);

    wprintf(curbwin->win, STAT_TEXT, 1, "\n");
  }
  return;
}
  
void conio_handleline(char *line) {

  /*  if(mfperl_handle(line) == 0)
      return; */
  if (*line == '/') {
    if(conio_handlecmd(line+1) == 0) {
      if (curbwin == &nullbwin) {
	conio_openwin(line+1);
	return;
      }
    } else {
      return;
    }
  }

  if(curbwin == &nullbwin) {
    conio_sendim(atom(line), firstwhite(line));
    return;
  }

  if (curbwin->chat)
    conio_sendchatim(curbwin->winname, line);
  else
    conio_sendim(curbwin->winname, line);
  return;  
}


/* XXX: you might want to make sure I converted to MAXCMDLEN usage in
 * a logical way.  I'm not actually reading the code :)  -- mid 
 * Conversion rules:  1024 -> MAXCMDLEN+1
 *                    1023 -> MAXCMDLEN
 */
void gotkey(int c) {
  
#ifndef NO_HISTORY
  HISTORY_STATE	*hstate = NULL;
  static int	histpos;
#endif

  static int	bufloc, bufmark, bufbuf;
  static char	buf[MAXCMDLEN+1], killbuf[MAXCMDLEN+1], inbold, initalic;

  updateidletime();

  if (c == 0) {
    bzero(buf, MAXCMDLEN+1);
    bzero(killbuf, MAXCMDLEN+1);
    bufloc = bufmark = initalic = inbold = 0;
#ifndef NO_HISTORY
    histpos = 0;
#endif
    return;
  }
  bufbuf = bufloc;

  switch(c) {
  case ERR:
    break;
    
    /* *******************************
       HISTORY STUFF
       ****************************** */
  case KEY_UP:
#ifndef NO_HISTORY
    hstate = history_get_history_state();
    if (hstate->entries == NULL)
      break;
    if (histpos > 0) {
      histpos--;
      assert(hstate->entries[histpos] != NULL);
      bzero(buf, MAXCMDLEN+1);
      strncpy(buf, (hstate->entries[histpos])->line, MAXCMDLEN);
      //      werase(win_input);
      //      wprintf(win_input, INPUT_TEXT, 1, "%s", buf);
      bufloc = strlen(buf);
    }
#endif
    break;
  case KEY_DOWN:
#ifndef NO_HISTORY
    hstate = history_get_history_state();
    assert(hstate != NULL);
    if (hstate->entries == NULL)
      break;
    if (histpos < hstate->length) {
      histpos++;
      bzero(buf, MAXCMDLEN+1);
      if (hstate->entries[histpos] != NULL)
	strncpy(buf, (hstate->entries[histpos])->line, MAXCMDLEN);
      //      werase(win_input);
      //winprintf(win_input, INPUT_TEXT, 1, "%s", buf);
      bufloc = strlen(buf);
    }
#endif
    break;

    /* ******************************
       LINE EDITING
       ****************************** */

  case KEY_LEFT:
    if (bufloc > 0) {
      bufloc--;
    }
    break;
  case KEY_RIGHT:
    if (buf[bufloc] != 0)
      bufloc++;
    break;

#if USE_LINEEDIT    
  case 1: case KEY_HOME: /* ^A , C-a , beginning of line */
    bufmark = bufloc;
    bufloc = 0;
    wclear(win_input);
    break;

  case 5: case KEY_END: /* ^E , C-e , end of line */
    bufmark = bufloc;
    bufloc = strlen(buf);
    if((bufloc % WINPUT_X) < (bufmark % WINPUT_X))
      wclear(win_input);
    break;

  case KEY_IC: /* toggle overwrite/insert */
    /* XXX : seems to have some "issues" */
    input_over = (input_over == 0);
    if(input_over)
      printf("Now in OVERWRITE mode\n");
    else
      printf("Now in INSERT mode\n");
    break;

  case 11: /* ^K , C-k , kill to end of line */
    bufmark = strlen(buf);
    conio_whomp(killbuf,buf,bufloc,bufmark,MAXCMDLEN+1);
    wclear(win_input);
    break;

  case 21: /* ^U , C-u , kill to beginning of line */
    if(bufloc == 0)
      break;

    conio_whomp(killbuf,buf,0,bufloc,MAXCMDLEN+1);
    bufloc = bufmark = 0;
    break;

  case 23: /* ^W , C-w , whomp */
    bzero(killbuf,MAXCMDLEN+1);
    conio_whomp(killbuf,buf,bufloc,bufmark,MAXCMDLEN+1);
    break;

  case 24: /* ^X , C-x , set mark */
    bufmark = bufloc;
    break;

  case 25: /* ^Y , C-y , yank */
    bufloc = conio_yank(buf, killbuf, bufloc);
    break;



#endif /* USE_LINEEDIT */

  case '\b': case KEY_BACKSPACE:
    if (bufloc < 1)
      break;
    bufloc--;
    if(buf[bufloc+1] == 0)
      buf[bufloc] = 0;
    else 
      memmove( (buf+bufloc),(buf+bufloc+1),(MAXCMDLEN+1-bufloc));
    break;
    
    
    /* ********************************
       WINDOW MOVING
       ******************************** */

  case KEY_PPAGE: case KEY_BTAB:
    curbwin = curbwin->prev;
    wbuddy_update();
    touchwin(curbwin->win);
    break;

  case '\t': case KEY_NPAGE:
    curbwin = curbwin->next;
    wbuddy_update();
    touchwin(curbwin->win);
    break;

  case 4: /* C-d ^D logout */
    conio_exit(NULL);
    break;

    /* ***************************************
       SPECIAL CHARACTERS
       *************************************** */

  case '<':
#if USE_LINEEDIT
    bufloc = conio_yank(buf, "&lt;", bufloc);
#else
    buf[bufloc++] = '&';
    buf[bufloc++] = 'l';
    buf[bufloc++] = 't';
    buf[bufloc++] = ';';
#endif
    break;

  case '>':
#if USE_LINEEDIT
    bufloc = conio_yank(buf, "&gt;", bufloc);
#else
    buf[bufloc++] = '&';
    buf[bufloc++] = 'g';
    buf[bufloc++] = 't';
    buf[bufloc++] = ';';
#endif
    break;

  case '&':
#if USE_LINEEDIT
    bufloc = conio_yank(buf, "&amp;", bufloc);
#else
    buf[bufloc++] = '&';
    buf[bufloc++] = 'a';
    buf[bufloc++] = 'm';
    buf[bufloc++] = 'p';
    buf[bufloc++] = ';';
#endif
    break;

    /* ********************************************
       FUNCTION KEYS
       ******************************************** */

  case KEY_F(1):
    help_printhelp(NULL);
    break;

  case KEY_F(2): case 2:
#if USE_LINEEDIT
    if(inbold)
      bufloc = conio_yank(buf, "</B>", bufloc);
    else
      bufloc = conio_yank(buf, "<B>", bufloc);
#else
    buf[bufloc++] = '<';
    if (inbold)
      buf[bufloc++] = '/';
    buf[bufloc++] = 'B';
    buf[bufloc++] = '>';
#endif
    inbold = (inbold == 0);
    break;

  case KEY_F(3): case 22:
#if USE_LINEEDIT
    if(initalic)
      bufloc = conio_yank(buf, "</I>", bufloc);
    else
      bufloc = conio_yank(buf, "<I>", bufloc);
#else
    buf[bufloc++] = '<';
    if (initalic)
      buf[bufloc++] = '/';
    buf[bufloc++] = 'I';
    buf[bufloc++] = '>';
#endif
    initalic = (initalic == 0);
    break;

  case KEY_F(4):
    break;

  case KEY_F(5):
#if USE_LINEEDIT
    bufloc = conio_yank(buf, "&", bufloc);
#else
    buf[bufloc++] = '&';
#endif
    break;

  case KEY_F(6): case 12:
    werase(curbwin->win);
    break;

  case KEY_F(7): case 18: 
    conio_reload(NULL);
    break;

  case KEY_F(8): case 14:
    conio_jump(NULL);
    break;

  case KEY_F(9):
    conio_away(NULL);
    break;

  case KEY_F(10): 
    conio_closewin(NULL);
    break;

  case KEY_F(11):
#if USE_LINEEDIT
    bufloc = conio_yank(buf, "<", bufloc);
#else
    buf[bufloc++] = '<';
#endif
    break;

  case KEY_F(12):
#if USE_LINEEDIT
    bufloc = conio_yank(buf, ">", bufloc);
#else
    buf[bufloc++] = '>';
#endif
    break;

    /* *******************************************
       DEFAULT
       ******************************************* */

  case '\r': case '\n':
    if(secs_getvar_int("autounaway"))
      if(nullbwin.away)
	unsetaway();

    if ((buf[0] == 0) || (buf == NULL))
      break;
#ifndef NO_HISTORY
    add_history(buf);
#endif
    conio_handleline(buf);
#ifndef NO_HISTORY
    hstate = history_get_history_state();
    histpos = hstate->length;
#endif
    bzero(buf, MAXCMDLEN+1);
    bufloc = bufmark = inbold = initalic = bufloc = 0;
    werase(win_input);
    break;

  default:
    if (isprint(c)) {
      if (bufloc < MAXCMDLEN) {
	if(buf[bufloc] == 0 || input_over == 1) {
	  buf[bufloc] = c;
	} else {
	  memmove( (buf+bufloc+1), (buf+bufloc), MAXCMDLEN-bufloc);
	  buf[bufloc] = c;
	}
      }
      bufloc++;
    }
    break;
  } /* switch(c) {*/

  conio_print_input(buf, bufloc, bufbuf);
  return;
}

#if USE_LINEEDIT
void conio_whomp(char *dest, char *src, int start, int end, int n) {
  int l;

  if ( (dest == NULL) || (src == NULL) )
    return;

  if(start > end) {
    l = start;
    start = end;
    end = l;
  }

  l = strlen(src);

  if( (end - start) > n)
    end = start + n;

  bzero(dest, n);
  memcpy(dest, src+start, end-start);
  memmove(src+start, src+end, l-end);
  bzero(src+start+(l-end),end-start);
  return;
}

int conio_yank(char *destbuf, char *killbuf, int bufloc) {
  if(destbuf[bufloc] != 0)
    memmove((destbuf+bufloc+strlen(killbuf)),destbuf+bufloc,MAXCMDLEN-strlen(killbuf)-bufloc);//whee!    
  memcpy((destbuf+bufloc),killbuf,strlen(killbuf));
  bufloc += strlen(killbuf);
  return(bufloc);
}

#endif

void conio_print_input(char *buf, int bufloc, int bufbuf) {
  chtype *outbuf;
  int maxout = ((WINPUT_Y * WINPUT_X)-1), 
    count = 0, 
    startpos = 0;

  if (bufbuf == bufloc)
    return;

  if ( ((bufloc % WINPUT_X) > (bufbuf % WINPUT_X)) || (bufloc < bufbuf) || (buf[bufloc+1] != 0)) {

    outbuf = (chtype *)malloc((maxout+1) * sizeof(chtype));
    
    if(outbuf == NULL)
      printf("uhh.. couldn't allocate input window buffer. bummer dude.\n");
    
    bzero(outbuf, ((maxout+1) * sizeof(chtype)) );

    if(bufloc >= WINPUT_X)
      startpos = (bufloc - (bufloc % WINPUT_X) - WINPUT_X);
    else
      startpos = 0;
    
    for(count = 0; count < maxout; count++)
      outbuf[count] = buf[startpos+count];

    if(outbuf[bufloc-startpos] == 0)
      outbuf[bufloc-startpos] = ' ';

    outbuf[bufloc-startpos] |= A_UNDERLINE;
 
    werase(win_input);
    
    for(count = 0; ((count < maxout) && (outbuf[count] != 0)); count++)
      waddch(win_input, outbuf[count]);

    free(outbuf);
    wrefresh(win_input); 
  } else {
    waddstr(win_input, "\b \b");
    for(count = bufbuf; count < bufloc; count++)
      waddch(win_input,buf[count]);
    waddch(win_input, (' ' | A_UNDERLINE));
  } 
  return;
}
